from .triggon import Triggon
from .trig_func import TrigFunc
from ._internal._exceptions import (
    InvalidArgumentError, 
    MissingLabelError, 
)

__version__ = "0.1.0b4"

__all__ = [
    "Triggon", "TrigFunc", 
    "InvalidArgumentError", 
    "MissingLabelError",
]




import inspect
from typing import Any

from ._exceptions import (
    _FrameAccessError,
    _UnsetExitError,
    SYMBOL,
)


def _store_org_value(self, label: str, index: int, org_value: Any) -> None:
    target_index = self._org_value[label][index] 

    if target_index is None:
        self._org_value[label][index] = []

    self._org_value[label][index].append(org_value)

def _update_var_value(
    self, var_ref: tuple[str, ...] | list[tuple[str, ...]], 
    label: str, index: int, update_value: Any, 
    inner_index: int=None, to_org: bool=False,
) -> None:
    # var_ref can be:
    # - (file name, lineno, var name) for globals
    # - (file name, lineno, attr name, instance) for class attributes
    # - a list of either form

    if to_org:
        trig_flag = False
    else:
        trig_flag = True
  
    if len(var_ref) == 4:     
        setattr(var_ref[3], var_ref[2], update_value)
    else:    
        self._frame.f_globals[var_ref[2]] = update_value

    if self.debug:
        self._print_var_debug(
            label, index, trig_flag, inner_index, update_value,
        )


def _check_label_flag(self, label: str, cond: str | None) -> None:
    target_func = "set_trigger"

    name = label.lstrip(SYMBOL)
    self._check_exist_label(name)

    if self._disable_label[name] or self._trigger_flag[name]:
        return
    
    if cond is not None:
        self._get_target_frame(target_func)

        if not self._ensure_safe_cond(cond):
            return
        
    self._trigger_flag[name] = True

    if self.debug:
        self._get_target_frame(target_func)
        self._print_flag_debug(name, "active", clear=False)

    self._label_has_var(name, target_func)

def _label_has_var(
    self, label: str, called_func: str, to_org: bool=False,
) -> None:
    if self._var_list[label] is None:
        return
    
    if to_org:
        update_value = self._org_value[label]
    else:
        # Each index always holds a single value.
        update_value = self._new_value[label]

    self._get_target_frame(called_func)

    # Restore all variables to their original or new values 
    for i in range(len(self._var_list[label])):
        var_ref = self._var_list[label][i]

        if var_ref is None:
            continue
        elif isinstance(var_ref, list):
            if to_org:
                for i_2, val in enumerate(update_value[i]):
                    self._update_var_value(
                        var_ref[i_2], label, i, val, i_2, to_org,
                    )
            else:
                for i_2, val in enumerate(var_ref):
                    self._update_var_value(
                        val, label, i, update_value[i], i_2, to_org,
                    )             
        else:
            if to_org:
                self._update_var_value(
                    var_ref, label, i, update_value[i][0], to_org=to_org,
                )
            else:
                self._update_var_value(
                    var_ref, label, i, update_value[i], to_org=to_org,
                )

def _get_target_frame(
        self, target_name: str | list[str, str], has_exit: bool=False,
) -> None:
   if self._frame is not None:
      return

   frame = inspect.currentframe()

   while frame:
      if has_exit:
          if frame.f_code.co_name == "<module>":
              raise _UnsetExitError()
          elif frame.f_code.co_name == target_name:
              break
      elif isinstance(target_name, list):
          if frame.f_code.co_name in target_name:
              self._frame = frame.f_back
              break
      elif frame.f_code.co_name == target_name:
         self._frame = frame.f_back
         break 
      
      frame = frame.f_back

   if has_exit:
       return
   elif self._frame is None:
       raise _FrameAccessError()
   
   self._get_trace_info()

def _get_trace_info(self) -> None:
    if self._lineno is None:
        self._lineno = self._frame.f_lineno

    if self._file_name is None:
        self._file_name = self._frame.f_code.co_filename

def _clear_frame(self) -> None:
   # to prevent memory leak by releasing the frame reference
   self._frame = None
   self._lineno = None
   self._file_name = None





import ast
from typing import Any

from ._exceptions import (
   InvalidArgumentError, 
   MissingLabelError,
   SYMBOL,
) 
from ._var_analysis import VALUE_TYPE_ERROR


INVALID_LABEL = "Label cannot be empty or blank: `{label}`"
LABEL_TYPE_ERROR = "Label must be a string."


def _check_exist_label(self, label: str) -> None:
    try:
        self._new_value[label]
    except KeyError:
        raise MissingLabelError(label)
    
def _try_search_var(
      self, node: ast.Name | ast.Attribute, 
      err_check=False, get_full_name: bool=False,
) -> tuple[str, Any] | str:
   if isinstance(node, ast.Name):
      try:
         ret_value = self._frame.f_locals[node.id]
      except KeyError:
         ret_value = self._frame.f_globals[node.id]

      var_name = node.id 

      if get_full_name:
         return (var_name, var_name)  
   else:
      try:
         ret_value = self._frame.f_locals[node.value.id]
         var_name = node.value.id
      except KeyError:
         ret_value = self._frame.f_globals[node.value.id]
         var_name = node.value.id
      except AttributeError:
         if err_check:
            raise InvalidArgumentError(VALUE_TYPE_ERROR)
         
         (attr_id, attr_full_name) = _get_attr_fullname(node)

         try:
            ret_value = self._frame.f_locals[attr_id]
         except KeyError:
            ret_value = self._frame.f_globals[attr_id]
            
         var_name = attr_id

         if get_full_name:
            return (var_name, attr_full_name)
         
      if get_full_name:
         return (var_name, f"{var_name}.{node.attr}")

      if err_check:
         return var_name
      
   return (var_name, ret_value)
   

def _get_attr_fullname(node: ast.Attribute) -> tuple[str, str]:
   name = []

   while isinstance(node, ast.Attribute):
      name.append(node.attr)
      node = node.value

   if isinstance(node, ast.Name):
      name.append(node.id)
      name.reverse()
   else:
      raise TypeError(
         "The attribute structure used in `cond` is not supported."
      )

   return (name[0], ".".join(name))

def _get_nested_value(self, attr_name: str) -> str | int:
   parts = attr_name.split(".")

   try:
      ref = self._frame.f_locals[parts[0]]
   except KeyError:
      try:
         ref = self._frame.f_globals[parts[0]]
      except KeyError:
         raise InvalidArgumentError(
            f"`{attr_name}` is not defined in the current scope."
         )
      
   for attr in parts[1:]:
      ref = getattr(ref, attr)

   return ref

def _compare_value_counts(
      self, label: str, needed_indices_count: int,
) -> None:
   # Note: `len(total_values)` is always one more than the max index.

   if not isinstance(needed_indices_count, int):
      raise InvalidArgumentError("`index` must be a integer.")

   if len(self._new_value[label]) < needed_indices_count + 1:
      raise ValueError(
         f"Please set the new value for index {needed_indices_count}."
      )
   
def _count_symbol(self, label: str) -> int:
   if not isinstance(label, str):
      raise InvalidArgumentError(LABEL_TYPE_ERROR)
   
   count = 0

   # count only prefix symbols
   for symbol in label:
      if symbol == SYMBOL:
         count += 1
         continue
      return count
    

def _check_label_type(label: Any, allow_dict: bool=True) -> None:
   if not allow_dict and isinstance(label, dict):
      raise InvalidArgumentError(
         "`dict` type is not allowed for `label` in this function."
      )
   elif isinstance(label, (list, tuple)):
      return
   elif not isinstance(label, str):
      raise InvalidArgumentError(LABEL_TYPE_ERROR)
   elif label == "" or " " in label:
      raise InvalidArgumentError(INVALID_LABEL.format(label=label))


def _handle_arg_types(
      label: Any, value: Any | None, index: int=None,
) -> dict[str, Any]:
   # check `index` type
   if index is not None and not isinstance(index, int):
      raise InvalidArgumentError(
         "The `index` keyword argument must be an `int`."
      )
      
   # single label
   if isinstance(label, str):    
      if label == "" or " " in label:
         raise InvalidArgumentError(INVALID_LABEL.format(label=key))  
      elif isinstance(value, (list, tuple)):          
        return {label: value}    
      else:        
         return {label: value}   
   elif isinstance(label, dict):
      # multiple labels
      if len(label) == 0:
         raise InvalidArgumentError(
            "Please provide both a key and a value for the dictionary."
         )   
      elif value is None:
         for key in label.keys():
            if not isinstance(key, str):
               raise InvalidArgumentError(LABEL_TYPE_ERROR)
            elif key == "" or " " in key:
               raise InvalidArgumentError(INVALID_LABEL.format(label=key))
         
         return label
      
      raise InvalidArgumentError(
         "All required values must be included in the dictionary provided to `label` "
         "Do not pass them as separate arguments."
      )
   
   else:
      raise InvalidArgumentError("`label` must be a string or a dictionary.")


from typing import Any
import os

from ._sentinel import _no_value


LOG_DEBUG = "【Debug】 [File]: {file} [Line]: {lineno}, [Function]: `{func}`, "

LOG_FLAG = LOG_DEBUG + "[State]: Label `{label}` is {state}"

LOG_TRIG = LOG_DEBUG + "[Label]: `{label}`, [Action]: {action}"

LOG_VALUE = (
  LOG_DEBUG 
  + "[Label]: `{label}` (index {num}), [Flag]: {state}, [Value]: {value}"
)

LOG_VAR = (
  LOG_DEBUG 
  + "[Label]: `{label}` (index {num}), "
  "[Flag]: {state}, [Variable]: `{var}`, [Value]: {value}"
)


def _get_debug_info(self) -> str:
   func_name = self._frame.f_code.co_name

   if self._lineno is None:
      self._lineno = self._frame.f_lineno
   
   if self._file_name is None:
      self._file_name = self._frame.f_code.co_filename

   return func_name

def _get_var_info(
        self, label: str, index: int, inner_index: int=None,
) -> tuple[str, Any]:
   if inner_index is None:
       var_refs = self._var_list[label][index]
       org_value = self._org_value[label][index][0]
   else:
       var_refs = self._var_list[label][index][inner_index]
       org_value = self._org_value[label][index][inner_index]      
   

   return (var_refs[2],org_value)

def _print_flag_debug(
      self, label: str, state: str, clear: bool=True,
) -> None:
    func_name = self._get_debug_info()
    file = os.path.basename(self._file_name)

    print(
        LOG_FLAG.format(
           file=file, lineno=self._lineno, 
           func=func_name, label=label, state=state,
        )
    )

    if clear:
       self._clear_frame()


# for `switch_lit()`

def _print_val_debug(
      self, 
      label: str | list[str] | tuple[str, ...], index: int, 
      flag: bool, before: Any, after: Any=_no_value,
) -> None:
      func_name = self._get_debug_info()
      file = os.path.basename(self._file_name)

      if after is _no_value:
         print(
            LOG_VALUE.format(
               file=file, lineno=self._lineno, 
               func=func_name, label=label, num=index, 
               state=flag, value=repr(before),
            )
         )
      else:
         print(
            LOG_VALUE.format(
               file=file, lineno=self._lineno, 
               func=func_name, label=label, num=index, 
               state=flag, value=f"{repr(before)} -> {repr(after)}",
            )
         )

      self._clear_frame()


# for `switch_var()`

def _print_var_debug(
      self, label: str, index: int, flag: bool, 
      inner_index: int=None, new_value: Any=_no_value,
) -> None:
   func_name = self._get_debug_info()
   file = os.path.basename(self._file_name)

   # `var_value` is the original value of this variable
   (var_name, var_value) = self._get_var_info(label, index, inner_index)

   if new_value is _no_value:
         value = repr(var_value)
   elif type(var_value) != type(new_value) or var_value != new_value:
         value = f"{repr(var_value)} -> {repr(new_value)}"
   else:
         new = self._new_value[label][index]
         value = f"{repr(new)} -> {repr(new_value)}"

   print(
      LOG_VAR.format(
         file=file, lineno=self._lineno, 
         func=func_name, label=label, num=index, 
         state=flag, var=var_name, value=value,
      )
   )


# for `set_trigger()` and `revert()`

def _print_trig_debug(self, label: str, action: str) -> None:
    func_name = self._get_debug_info()
    file = os.path.basename(self._file_name)

    print(
        LOG_TRIG.format(
           file=file, lineno=self._lineno, func=func_name, 
           label=label, action=action,
        )
    )
    self._clear_frame()



def _init_or_not(self, label: str, index: int) -> bool:
    self._get_target_frame(["switch_var", "alter_var"]) # Will change it after beta 

    self._check_exist_label(label)  
    self._compare_value_counts(label, index)

    if (
        self._var_list[label][index] is None 
        or self._is_new_var(label, index)
    ):  
        # Initial process to store argument variables 
        self._trace_func_call()
        return True
    
    return False

def _is_new_var(
        self, label: str, index: int, target_ref: tuple[str, ...]=None,
) -> bool:
    # If both the file name and line number match,
    # the variable is already registered.

    var_ref = self._var_list[label][index]

    if isinstance(var_ref, tuple):
        if target_ref is not None:
            if target_ref[:3] == var_ref[:3]:
                return False
            else:
                return True
        elif var_ref[0] == self._file_name and var_ref[1] == self._lineno:
            return False
        
        return True
    
    for ref in var_ref:
        if target_ref is not None:
            if target_ref[:3] == var_ref[:3]:
                return False
        elif ref[0] == self._file_name and ref[1] == self._lineno:
            return False
        
    return True

def _find_match_var(self, label: str=None, index: int=None) -> None:
    # Used to retrieve the original value for `revert()`

    # it can be a list or tuple
    var_ref = self._var_list[label][index]

    # Searrch up to the same label
    stop_flag = False

    # `value` is a list
    for key, value in self._var_list.items(): 
        if stop_flag:
            return
        
        if key == label:
            stop_flag = True

        # `list_val` is a list or tuple or None
        for i, list_val in enumerate(value):
            if list_val is None:
                continue
            elif isinstance(list_val, tuple) and isinstance(var_ref, tuple):
                if self._is_ref_match(list_val, var_ref):
                    self._org_value[label][index] = self._org_value[key][i]     
            elif isinstance(list_val, tuple) and isinstance(var_ref, list):
                for i_2, ref_v in enumerate(var_ref):
                    if self._is_ref_match(list_val, ref_v):
                        self._org_value[label][index][i_2] = (
                            self._org_value[key][i]
                        )
            elif isinstance(list_val, list) and isinstance(var_ref, tuple):
                for i_2, l_v in enumerate(list_val):
                    if self._is_ref_match(l_v, var_ref):
                        self._org_value[label][index] = (
                            self._org_value[key][i][i_2]
                        )
            else:
                for ref_i, ref_val in enumerate(var_ref):
                    for list_i, v in enumerate(list_val):
                        if self._is_ref_match(v, ref_val):                           
                            self._org_value[label][index][ref_i] = (
                                self._org_value[key][i][list_i]
                            )

def _is_ref_match(
        self, list_val: tuple[str, ...], target_val: tuple[str, ...],
) -> bool:
    if len(list_val) == 3 and len(target_val) == 3:
        return list_val[0] == target_val[0] and list_val[2] == target_val[2]
    elif len(list_val) == 4 and len(target_val) == 4:
        return list_val[2:] == target_val[2:]
    


import ast
import linecache
from itertools import count

from ._exceptions import (
   InvalidArgumentError,
   SYMBOL, 
)


LABEL_ERROR = "Label must be a string."
VALUE_TYPE_ERROR = "Provided value is not a variable."
VAR_ERROR = "Local arguments are not supported in this function."
NEST_ERROR = (
   "Variables should not be nested in lists or tuples (e.g. [x, [y]])."
)


def _trace_func_call(self) -> None:
    lines = ""
    
    for i in count(self._lineno):
      line = linecache.getline(self._file_name, i)

      if not line:
        break
      lines += line

      try:
        line_range = ast.parse(lines.lstrip()) 
      except SyntaxError:
        # Skip lines where the code is incomplete,
        # such as unfinished multi-line function calls,
        # or lines that would raise an `IndentationError`.
        continue
      else:
        # walk through AST nodes to find function calls
        
        for node in ast.walk(line_range):
          if not isinstance(node, ast.Call):
            continue    

          if (
             not isinstance(node.func, ast.Attribute) 
             or node.func.attr not in ["switch_var", "alter_var"]
          ): 
            continue

          first_arg = node.args[0]

          # handle each argument type differently

          if isinstance(first_arg, ast.Dict):
            result = self._arg_is_dict(first_arg)
          else:
             try:
               second_arg = node.args[1]
             except IndexError:
                raise InvalidArgumentError(
                   "Provide variables as the second argument."
                )
             
             _identify_arg(second_arg)

             label = self._try_search_value(first_arg)

             # check if 'index' keyword is set
             if 0 < len(node.keywords):
                index = None

                for kw in node.keywords:
                   if kw.arg != "index":
                      continue
                   
                   index_node = kw.value
                   index = self._try_search_value(index_node)

                if index is None:
                   index = self._count_symbol(label)
             else:
                index = self._count_symbol(label)

             name = label.lstrip(SYMBOL) 

             self._check_exist_label(name)
             self._compare_value_counts(name, index)

             if isinstance(second_arg, (ast.List, ast.Tuple)):
                result = self._arg_is_seq(second_arg, name, index)
             elif isinstance(second_arg, ast.Name):
                result = self._arg_is_name(second_arg.id, name, index)
             elif isinstance(second_arg, ast.Attribute):
                result = self._arg_is_attr(second_arg, name, index)
             else:
                raise InvalidArgumentError(VALUE_TYPE_ERROR)

        # It means that the target function was found at least once.
        if result:
           return

        raise RuntimeError(
           "Call to 'alter_var' or 'switch_var' could not be detected. "
           "This may occur in certain environments " 
           "where the source code is unavailable or dynamically executed."
        )   
  
def _arg_is_dict(self, target: ast.Dict) -> bool:
    arg_list = self._deduplicate_labels(target)
  
    for key, val in arg_list.items():
      label = key.lstrip(SYMBOL)
      index = self._count_symbol(key)

      self._check_exist_label(label)
      self._compare_value_counts(label, index)

      # Handle each argument type differently

      if isinstance(val, (ast.List, ast.Tuple)):
        _ = self._arg_is_seq(val, label, index)
      elif isinstance(val, ast.Dict):
         raise InvalidArgumentError(VALUE_TYPE_ERROR)
      else:
        _identify_arg(val)

        if isinstance(val, ast.Name):
          _ = self._arg_is_name(val.id, label, index)
        elif isinstance(val, ast.Attribute):
          _ = self._arg_is_attr(val, label, index)
        else:
           raise InvalidArgumentError(VALUE_TYPE_ERROR)

    return True

def _arg_is_seq(
      self, target: ast.List | ast.Tuple, label: str, index: int,
) -> bool:
    target_index = self._var_list[label][index] 

    if target_index is None:
      self._var_list[label][index] = []
    elif isinstance(target_index, tuple):
       # convert to a list for adding values
       self._var_list[label][index] = [target_index]

    for val in target.elts:
      _identify_arg(val)

      if isinstance(val, ast.Name):
        _ = self._arg_is_name(val.id, label, index)
      elif isinstance(val, ast.Attribute):
        _ = self._arg_is_attr(val, label, index)  
      elif isinstance(val, (ast.List, ast.Tuple, ast.Dict)):
         raise InvalidArgumentError(NEST_ERROR)
      else:
        raise InvalidArgumentError(VALUE_TYPE_ERROR)

    return True

def _arg_is_name(self, target: str, label: str, index: int,) -> bool: 
   try:
      org_value = self._frame.f_globals[target]
   except KeyError:
      raise InvalidArgumentError(VAR_ERROR)

   target_index = self._var_list[label][index]

   # (file neme, line number, var name)
   var_refs = (self._file_name, self._lineno, target)

   if target_index is not None:
      if not self._is_new_var(label, index, var_refs):
         # The variable has already been registered
         return True
      
      if isinstance(target_index, tuple):
         self._var_list[label][index] = [target_index]

      self._var_list[label][index].append(var_refs)
   else:
      self._var_list[label][index] = var_refs

   self._store_org_value(label, index, org_value)
   self._find_match_var(label, index)

   return True

def _arg_is_attr(
      self, target: ast.Attribute, label: str, index: int,
) -> bool:
    var_name = self._try_search_var(target, err_check=True)
    instance = self._frame.f_locals[var_name]

    target_index = self._var_list[label][index]

    # (file name, line number, attr name, class instance)
    var_refs = (self._file_name, self._lineno, target.attr, instance)

    if target_index is not None:
      if not self._is_new_var(label, index, var_refs):
         # The variable has already been registered
         return True
      
      if isinstance(target_index, tuple):
         self._var_list[label][index] = [target_index]

      self._var_list[label][index].append(var_refs)
    else:
       self._var_list[label][index] = var_refs

    org_value = instance.__dict__[target.attr]  
    self._store_org_value(label, index, org_value)
    self._find_match_var(label, index)

    return True 
    
def _ensure_safe_cond(self, expr: str) -> bool:
   if not isinstance(expr, str):
      raise InvalidArgumentError("'cond' must be a string if provided.")
   
   allowed = (
        ast.Expression,
        ast.Compare, ast.Name, ast.Attribute, ast.Constant, 
        ast.Is, ast.IsNot, ast.In, ast.NotIn,
        ast.Load, ast.Eq, ast.NotEq, ast.Lt, 
        ast.Gt, ast.LtE, ast.GtE, ast.BoolOp, 
        ast.And, ast.Or, ast.UnaryOp, ast.Not,
   )

   try:
      tree = ast.parse(expr, mode="eval")
   except SyntaxError:
      raise InvalidArgumentError(
         "The expression passed to `cond` has a syntax error."
      )
   
   scope = {}

   for node in ast.walk(tree):
      if not isinstance(node, allowed):
         raise InvalidArgumentError(
            "Only comparison expressions "
            "(e.g. `x > 10`, `a == b`) are allowed."
         )
      
      if isinstance(node, ast.Name) or isinstance(node, ast.Attribute):
         (var_name, var_value) = self._try_search_var(node)
         scope[var_name] = var_value
    
   return eval(expr, scope)      
   
def _deduplicate_labels(self, target: ast.Dict) -> dict[str, ast.AST]:
    # Remove duplicate labels by converting to a set
    sorted_list = {}

    for key, val in zip(target.keys, target.values):
      label = self._try_search_value(key) 
      sorted_list[label] = val
         
    return sorted_list

def _try_search_value(self, var: ast.AST) -> int | str:
   if isinstance(var, ast.Constant):
      value = var.value
   elif isinstance(var, ast.Name):
      (_, value) = self._try_search_var(var)
   elif isinstance(var, ast.Attribute):
      (var_name, full_name) = self._try_search_var(var, get_full_name=True)

      if full_name.count(".") > 1:
         # When it's an attribute chain
         value = self._get_nested_value(full_name)
      else:
         instance = self._frame.f_locals[var_name]
         value = instance.__dict__[var.attr]
   else:
      raise InvalidArgumentError(
         "Only literals, variables, or simple attribute access " 
         "are allowed for labels or the index keyword in this function."
      )
   
   return value


def _identify_arg(target: ast.AST) -> None:
    if isinstance(target, ast.Constant):
      raise InvalidArgumentError(VALUE_TYPE_ERROR)
    
      



from types import FrameType
from typing import Any

from .trig_func import TrigFunc
from ._internal import (
  _debug,
  _err_handler,
  _var_analysis,
  _var_update,
  _switch_var,
)
from ._internal._err_handler import (
  _check_label_type,
  _handle_arg_types, 
)
from ._internal._exceptions import (
  InvalidArgumentError,
  SYMBOL,
  _ExitEarly,
)
from ._internal._sentinel import _no_value


class Triggon:
    debug: bool
    _debug_var: dict[str, tuple[int, str] | list[tuple[int, str]]]
    _trigger_flag: dict[str, bool]
    _new_value: dict[str, tuple[Any, ...]]
    _org_value: dict[str, list[Any]]
    _var_list: dict[str, tuple[str, ...] | list[tuple[str, ...]]]
    _disable_label: dict[str, bool]
    _return_value: tuple[bool, Any] | None
    _file_name: str
    _lineno: int
    _frame: FrameType

    # `new_value`: Each label holds a tuple of values.
    # `org_value`: Each label index holds a list of values,
    #              or None if unset.
    # `var_list`: Each label index holds a tuple of strings,
    #             a list of such tuples, or None if unset.

    def __init__(
        self, label: str | dict[str, Any], /, new: Any=None, 
        *, debug: bool=False,
    ) -> None:
      """
      Registers labels and their corresponding values.

      Values must be passed in the order of their index positions.
      To treat a sequence as a single value, wrap it in another sequence.

      Accepted formats:
      - label, value
      - label, [value]
      - label, (value,)
      - {label: value}
      - {label: [value]}
      - {label: (value,)}
      """

      self.debug = debug
      self._trigger_flag = {}
      self._new_value = {}
      self._org_value = {}
      self._var_list = {}  
      self._disable_label = {}   
      self._return_value = None
      self._file_name = None
      self._lineno = None
      self._frame = None

      change_list = _handle_arg_types(label, new)
      self._scan_dict(change_list)

    def _scan_dict(self, arg_dict: dict[str, Any]) -> None:      
      for key, value in arg_dict.items():          
          index = self._count_symbol(key)

          if index != 0:
              raise InvalidArgumentError(
                  f"Please remove the `*` prefix from `{key}`. " 
                  "To specify by index, "
                  "provide the values in index order using a list or tuple."
              )
          
          self._add_new_label(key, value)

    def _add_new_label(self, label: str, value: Any, /) -> None:
      # The value is normalized to a tuple
      if isinstance(value, (list, tuple)):
        length = len(value)

        if length == 0:
          # An empty sequence is handled as a single value
          length = 1
          self._new_value[label] = (value,)
        else:
          self._new_value[label] = tuple(value)
      else:
        length = 1
        self._new_value[label] = (value,)

      self._trigger_flag[label] = False
      self._disable_label[label] = False

      # Create a list of `None` values—one for each index of this label
      # (`length` is greater than 0)
      self._org_value[label] = [None] * length
      self._var_list[label] = [None] * length

    def set_trigger(
        self, label: str | list[str] | tuple[str, ...], /, *, cond: str=None,
    ) -> None:
      """
      Activates the trigger flag for the given labels.

      If variables were registered via `alter_var()`, 
      their values will be updated when the flag is activated.

      If `cond` is provided, it must be a valid comparison expression 
      (e.g., `"x > 10"`, `"obj.count == 5"`). 
      The expression is evaluated safely and the trigger will only be 
      activated if the result is `True`.
      """

      if isinstance(label, (list, tuple)):
        for name in label:
          _check_label_type(name, allow_dict=False)       
          self._check_label_flag(name, cond)
      else:
        _check_label_type(label, allow_dict=False)
        self._check_label_flag(label, cond)
        
      self._clear_frame()

    def switch_lit(
        self, label: str | list[str] | tuple[str, ...], /, org: Any, 
        *, index: int=None,
    ) -> Any:
      """
      Changes the value at the specified label(s) and position 
      if the flag is active.

      Only accepts immediate values
      (e.g., literals or expressions).

      Note:
        If multiple labels share the same name,  
        all will be triggered together by `set_trigger()`, 
        regardless of index.  
        In such cases, the one with the smaller index in the sequence takes precedence.  
        This also applies when different labels are active at once.
      """

      cur_functions = ["switch_lit", "alter_literal"] # Will change it after beta

      _check_label_type(label, allow_dict=False)

      if isinstance(label, (list, tuple)):
        for v in label:
          stripped_label = v.lstrip(SYMBOL)
          self._check_exist_label(stripped_label)

          if self._trigger_flag[stripped_label]:
            label = v
            break

        # When all labels' triggers are inactive
        if not isinstance(label, str):
          return org

      name = label.lstrip(SYMBOL)
      self._check_exist_label(name)

      if index is None:
        index = self._count_symbol(label)
      self._compare_value_counts(name, index)

      flag = self._trigger_flag[name]

      if not flag:
        ret_value = org
        new_val = _no_value # for debug
      else:
        ret_value = self._new_value[name][index] 
        new_val = self._new_value[name][index] # for debug

      if self.debug:
        self._get_target_frame(cur_functions)
        self._print_val_debug(name, index, flag, org, new_val)

      return ret_value

    def switch_var(
          self, label: str | dict[str, Any], var: Any=None, /, 
          *, index: int=None,
    ) -> None | Any:
        """
        Change the value of variables associated with the given label 
        if the flag is active.

        Only supports variable references (not literals or expressions).

        Supports updating:
        - Global variables
        - Class attributes (fields)
        """

        change_list = _handle_arg_types(label, var, index)
        init_flag = False

        if len(change_list) == 1:
          # When only one label is provided
          label = next(iter(change_list))
          name = label.lstrip(SYMBOL)   

          if index is None:
            index = self._count_symbol(label)

          if not init_flag:
            init_flag = self._init_or_not(name, index)

          trig_flag = self._trigger_flag[name]
          vars = self._var_list[name][index]

          if not trig_flag:
            self._clear_frame()
            return var
          elif not init_flag:
             return var

          self._update_var_value(
            vars, name, index, self._new_value[name][index],
          )      
          self._clear_frame()

          return var
        else:
           # When multiple labels are provided in a dictionary

          if index is not None:
            raise InvalidArgumentError(
              "Cannot use the `index` keyword with a dictionary. " 
              "Use `*` in the label instead." 
            )
          
          for key in change_list.keys():
            name = key.lstrip(SYMBOL)
            index = self._count_symbol(key)

            if not init_flag:
              init_flag = self._init_or_not(name, index)
            
            if not init_flag:
              continue

            trig_flag = self._trigger_flag[name]
            vars = self._var_list[name][index]  

            if not trig_flag:
              continue          

            self._update_var_value(
              vars, name, index, self._new_value[name][index],
            )
            
          self._clear_frame()

    def revert(
          self, label: str | list[str] | tuple[str, ...]=None, /, 
          *, all: bool=False, disable: bool=False,
    ) -> None:
      """
      Revert the trigger flag(s) set by `set_trigger()` back to False.

      Use the `all` keyword to revert all labels at once. 
      If `disable` is set to True, the label(s) will be permanently disabled.
      """

      if label is None:
        if not all:
          raise InvalidArgumentError("No labels specified to revert.")
        
        for key in self._new_value.keys():
          self._revert_label(key, disable)      
      elif isinstance(label, (list, tuple)):
        for name in label:
          self._revert_label(name, disable)
      else:
        self._revert_label(label, disable)  

      self._clear_frame()      

    def _revert_label(self, label: str, disable: bool) -> None:
      _check_label_type(label, allow_dict=False)

      name = label.lstrip(SYMBOL)
      self._check_exist_label(name)

      if not disable and not self._trigger_flag[name]:
        return
      elif disable and self._disable_label[name]:
        return 

      if disable:
        state = "disable"  # for debug
        self._disable_label[name] = True
      else:
        state = "inactive" # for debug
      self._trigger_flag[name] = False

      self._label_has_var(name, "revert", to_org=True)

      if self.debug:
        self._get_target_frame("revert")
        self._print_flag_debug(name, state)    
    
    def exit_point(self, label: str, func: TrigFunc, /) -> None | Any:
      """
      Handles an early return triggered by `trigger_return()`,  
      based on the specified label.
      """

      name = label.lstrip(SYMBOL)
      self._check_exist_label(name)

      try:
          return func()
      except _ExitEarly:
          if not self._return_value[0]:
              return self._return_value[1]
          
          print(self._return_value[1])

    def trigger_return(
        self, label: str, /, ret: Any=None, 
        *, index: int=None, do_print: bool=False,
    ) -> None | Any:
        """
        Executes an early return using the set return value,  
        if the trigger flag is active.

        If `do_print` is True, prints the value with the early return.  
        Raises `InvalidArgumentError` if the value is not a string.
        """

        name = label.lstrip(SYMBOL)
        self._check_exist_label(name)

        if index is None:
           index = self._count_symbol(label)

        if not self._trigger_flag[name]:
            return 
            
        if do_print:
           if ret is None and not isinstance(self._new_value[name][index], str):
             raise InvalidArgumentError(
                "Expected a value of type `str`, "
                f"but got `{type(self._new_value[name][index]).__name__}`."
             )   
           elif ret is not None and not isinstance(ret, str):
             raise InvalidArgumentError(
                "Expected a value of type `str`, "
                f"but got `{type(ret).__name__}`."
             )

        if ret is None:
          value = self._new_value[name][index]
        else:
          value = ret   
                    
        self._return_value = (do_print, value)

        if self.debug:
          self._get_target_frame("trigger_return")
          self._print_trig_debug(name, "Return")

        self._get_target_frame("exit_point", has_exit=True)

        raise _ExitEarly 
        
    def trigger_func(self, label: str, func: TrigFunc, /) -> None | Any:
        """
        Executes the given function 
        if the trigger flag for the label is active.
        """

        name = label.lstrip(SYMBOL)
        self._check_exist_label(name)

        if self._trigger_flag[name]:
            if self.debug:
              self._get_target_frame("trigger_func")
              self._print_trig_debug(name, "Trigger a function") 
                    
            return func()
        
    # Old functions
    alter_literal = switch_lit
    alter_var = switch_var


modules = [_debug, _err_handler, _var_analysis, _var_update, _switch_var]

for module in modules:
  for name, func in vars(module).items():
      if callable(func):
          setattr(Triggon, name, func)


SYMBOL = "*"


class InvalidArgumentError(ValueError): 
    pass


class MissingLabelError(KeyError):
    def __init__(self, label: str):
        self.label = label

    def __str__(self):
        return f"Label `{self.label}` is missing."
    

class _UnsetExitError(Exception):
    def __init__(self):
        super().__init__(
            "Please set `exit_point()` before calling this function."
        )

class _FrameAccessError(Exception):
    def __init__(self):
        super().__init__(
            "Failed to get the current execution frame."
        )


class _ExitEarly(Exception): 
    pass

   